%{
/***************************************************************************
 *   Copyright (C) 2005 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <glib.h>		/* Just for g_free() */

#define YYSTYPE char *

void yyerror (char *msg);
void reset_parse_data (void);

/* From cmpfile.c */
void cmpfile_add_rom_callback (char *name, char *description, char *filename, char *filesize, char *crc32);

extern int yylineno;

char *last_game_name = NULL;
char *last_game_desc = NULL;
char *last_rom_name = NULL;
char *last_rom_size = NULL;
char *last_rom_crc = NULL;
char *last_rom_md5 = NULL;

/* http://www.logiqx.com/DatFAQs/CMPro.php */
%}

%token CLRMAMEPRO CATEGORY VERSION AUTHOR
	GAME NAME DESCRIPTION
	SIZE CRC MD5 ROM
	STRING

%start cmp_file

%%

cmp_file	: cmp_header game_entry_list
		;

cmp_header	: CLRMAMEPRO '(' header_opts ')'	{/*printf ("ClrMamePro header found\n");*/}
		;
		
header_opt	: name_tag				{cmpfile_set_name_callback ($$);
							 g_free ($$);}
		| description_tag			{cmpfile_set_description_callback ($$);
				 			 g_free ($$);}
		| category_tag				{cmpfile_set_category_callback ($$);
							 g_free ($$);}
		| version_tag				{cmpfile_set_version_callback ($$);
							 g_free ($$);}
		| author_tag				{cmpfile_set_author_callback ($$);
							 g_free ($$);}
		;
		
header_opts	: header_opts header_opt
		| header_opt
		;
		
category_tag	: CATEGORY STRING			{$$ = $2;}
		;
		
version_tag	: VERSION STRING			{$$ = $2;}
		;
		
author_tag	: AUTHOR STRING				{$$ = $2;}
		;
		
game_entry_list : game_entry_list game_entry
		| game_entry
		;
		
game_opt	: name_tag				{last_game_name = $$;}
		| description_tag			{last_game_desc = $$;}
		| rom_tag
		;
		
game_opts	: game_opts game_opt
		| game_opt
		;
		
game_entry	: GAME '(' game_opts ')'		{cmpfile_add_rom_callback (last_game_name, last_game_desc, last_rom_name, last_rom_size, last_rom_crc);
							 /*printf ("New game: %s/%s\n", last_game_name, last_rom_crc); */
							 reset_parse_data ();}
		;
		
name_tag	: NAME STRING				{$$ = $2;}
		;
		
description_tag	: DESCRIPTION STRING			{$$ = $2;}
		;
		
size_tag	: SIZE STRING				{$$ = $2;}
		;
		
crc_tag		: CRC STRING				{$$ = $2;}
		;
		
md5_tag		: MD5 STRING				{$$ = $2;}
		;

rom_opt		: name_tag				{last_rom_name = $$;}
		| size_tag				{last_rom_size = $$;}
		| crc_tag				{last_rom_crc = $$;}
		| md5_tag				{last_rom_md5 = $$;}
		;
		
rom_opts	: rom_opts rom_opt
		| rom_opt
		;
		
rom_tag		: ROM '(' rom_opts ')'
		;
		
%%

/* Parse error */
void yyerror (char *m) {
	printf ("Error at line %d: %s\n", yylineno, m);

	return;
}


void reset_parse_data (void) {
	g_free (last_game_name);
	g_free (last_game_desc);
	g_free (last_rom_name);
	g_free (last_rom_size);
	g_free (last_rom_crc);
	g_free (last_rom_md5);
	
	last_game_name = NULL;
	last_game_desc = NULL;
	last_rom_name = NULL;
	last_rom_size = NULL;
	last_rom_crc = NULL;
	last_rom_md5 = NULL;

	return;
}


// int main (void) {
// 	return (yyparse ());
// }
