/***************************************************************************
 *   Copyright (C) 2006 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include <stdio.h>
#include <glib.h>
#include <glib/gstdio.h>
#include "datfile.h"
#include "romentry.h"
#include "file.h"

/* Provided by the Bison-generated scanner */
extern int yyparse (void);

static datfile *out;

datfile *datfile_new_from_filename_cmp (char *filename) {
	GTimer *timer;
	extern FILE *yyin;
	int err;

	out = NULL;
	if ((yyin = fopen (filename, "r"))) {
		/* Init structure */
		out = g_new0 (datfile, 1);
		out -> filename = g_strdup (filename);
		g_datalist_init (&(out -> roms));
	
		timer = g_timer_new ();
		err = yyparse ();
		fclose (yyin);
		g_timer_stop (timer);
		
		if (err) {
			g_warning ("Error during parsing");
		} else {
			g_fprintf (stderr,
				"Filename..........: %s\n"
				"Read in...........: %lf seconds\n"
				"DAT format........: ClrMamePro\n"
				"ROMs..............: %d\n"
				"Author............: %s\n"
				"Data version......: %s\n"
				"Comment...........: %s\n"
				"Emulator..........: %s\n", filename, g_timer_elapsed (timer, NULL), out -> roms_no,
				out -> author, out -> data_version, out -> comment, out -> emulator);
		}
		g_timer_destroy (timer);
	}

	return (out);
}


void cmpfile_add_rom_callback (char *name, char *description, char *filename, char *filesize, char *crc32) {
	GQuark romquark;
	romentry *rom;

	rom = romentry_new (name, description, filename, filesize, crc32);
	romquark = g_quark_from_string (rom -> crc32);
	g_datalist_id_set_data (&(out -> roms), romquark, rom);
	(out -> roms_no)++;

	return;
}


/* The following functions follow the equivalency rules defined by Logiqx.
   See http://www.logiqx.com/DatFAQs/CMPro.php */

void cmpfile_set_name_callback (char *string) {
	out -> emulator = g_strdup (string);

	return;
}

void cmpfile_set_description_callback (char *string) {
	out -> emulator_version = g_strdup (string);

	return;
}

void cmpfile_set_category_callback (char *string) {
	out -> comment = g_strdup (string);

	return;
}


void cmpfile_set_version_callback (char *string) {
	out -> data_version = g_strdup (string);

	return;
}

void cmpfile_set_author_callback (char *string) {
	out -> author = g_strdup (string);

	return;
}
