/***************************************************************************
 *   Copyright (C) 2006 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <glib/gstdio.h>
#include "datfile.h"
#include "romentry.h"
#include "file.h"

/* ROMCenter file format: See http://www.logiqx.com/DatFAQs/RomCenter.php
parent nameparent descriptiongame namegame descriptionrom namerom crcrom sizeromof namemerge name
*/

datfile *datfile_new_from_filename_rc (char *filename) {
	GIOChannel *gc;
	GError *err;
	datfile *out;
	gchar *line, *section, **tokens;
	guint32 lineno;
	GTimer *timer;
	GQuark romquark;
	romentry *rom;

	err = NULL;
	if (!(gc = g_io_channel_new_file (filename, "r", &err))) {
		g_warning ("Cannot open RC file \"%s\": %s", filename, err -> message);
		g_error_free (err);
		out = NULL;
	} else {
// 		g_debug ("RC file opened OK");
		g_io_channel_set_encoding (gc, NULL, NULL);

		/* Init structure */
		out = g_new0 (datfile, 1);
		out -> filename = g_strdup (filename);
		g_datalist_init (&(out -> roms));
		
		section = NULL;
		timer = g_timer_new ();
		for (lineno = 1; g_io_channel_read_line (gc, &line, NULL, NULL, &err) == G_IO_STATUS_NORMAL; lineno++) {
			g_strstrip (line);
// 			g_print ("--> %s", line);
			if (g_pattern_match_simple ("[*]", line)) {
				if (section)
					g_free (section);
				section = g_strndup (line + 1, strlen (line) - 2);
// 				g_debug ("New section: \"%s\"", section);
			} else if (g_ascii_strcasecmp (section, "CREDITS") == 0) {
				tokens = g_strsplit (line, "=", 0);
				if (g_strv_length (tokens) != 2) {
					g_warning ("Malformed line at %d", lineno);
				} else {
					if (g_ascii_strcasecmp (tokens[0], "author") == 0) {
						out -> author = g_strdup (tokens[1]);
					} else if (g_ascii_strcasecmp (tokens[0], "version") == 0) {
						out -> data_version = g_strdup (tokens[1]);
					} else if (g_ascii_strcasecmp (tokens[0], "comment") == 0) {
						out -> comment = g_strdup (tokens[1]);
					} else {
// 						g_warning ("Unknown entry in CREDITS section at line %d: %s", lineno, tokens[0]);
					}
				}
				g_strfreev (tokens);
			} else if (g_ascii_strcasecmp (section, "DAT") == 0) {
				tokens = g_strsplit (line, "=", 0);
				if (g_strv_length (tokens) != 2) {
					g_warning ("Malformed line at %d", lineno);
				} else {
					if (g_ascii_strcasecmp (tokens[0], "version") == 0) {
						out -> dat_format = g_strdup (tokens[1]);
					} else if (g_ascii_strcasecmp (tokens[0], "plugin") == 0) {
						out -> plugin = g_strdup (tokens[1]);
					} else if (g_ascii_strcasecmp (tokens[0], "split") == 0) {
						sscanf (tokens[1], "%d", &(out -> split));
					} else if (g_ascii_strcasecmp (tokens[0], "merge") == 0) {
						sscanf (tokens[1], "%d", &(out -> merge));
					} else {
// 						g_warning ("Unknown entry in DAT section at line %d: %s", lineno, tokens[0]);
					}
				}
				g_strfreev (tokens);
			} else if (g_ascii_strcasecmp (section, "EMULATOR") == 0) {
				tokens = g_strsplit (line, "=", 0);
				if (g_strv_length (tokens) != 2) {
					g_warning ("Malformed line at %d", lineno);
				} else {
					if (g_ascii_strcasecmp (tokens[0], "refname") == 0) {
						out -> emulator = g_strdup (tokens[1]);
					} else if (g_ascii_strcasecmp (tokens[0], "version") == 0) {
						out -> emulator_version = g_strdup (tokens[1]);
					} else {
						g_warning ("Unknown entry in EMULATOR section at line %d: %s", lineno, tokens[0]);
					}
				}
				g_strfreev (tokens);
			} else if (g_ascii_strcasecmp (section, "GAMES") == 0) {
				/* New game */
				tokens = g_strsplit (line, "\xAC", 0);
				if (g_strv_length (tokens) != 10) {
					g_warning ("Malformed line at %d", lineno);
				} else {
					rom = romentry_new (tokens[3], tokens[4], tokens[5], tokens[7], tokens[6]);
					romquark = g_quark_from_string (rom -> crc32);
					g_datalist_id_set_data (&(out -> roms), romquark, rom);
					(out -> roms_no)++;
				}
				g_strfreev (tokens);
			}
		}
		g_timer_stop (timer);
		
		if (err) {
			g_warning ("Error: %s", err -> message);
			g_error_free (err);
		} else {
			g_fprintf (stderr,
				"Filename..........: %s\n"
				"Read in...........: %lf seconds\n"
				"DAT format........: RomCenter %s\n"
				"ROMs..............: %d\n"
				"Author............: %s\n"
				"Data version......: %s\n"
				"Comment...........: %s\n"
				"Emulator..........: %s\n", filename, g_timer_elapsed (timer, NULL), out -> dat_format,
			out -> roms_no, out -> author, out -> data_version, out -> comment, out -> emulator);
		}
		g_io_channel_unref (gc);		// = close()
		g_timer_destroy (timer);
	}

	return (out);
}


romentry *datfile_find (datfile *rcf, filez *f) {
	romentry *r;

	r = (romentry *) g_datalist_get_data (&(rcf -> roms), f -> crc32);

	return (r);
}


typedef enum {
	DATF_UNKNOWN,
	DATF_ROMCENTER,
	DATF_CLRMAMEPRO
} dat_format;

datfile *datfile_new_from_filename (char *filename) {
	GIOChannel *gc;
	GError *err;
	datfile *out;
	gchar *line;
	err = NULL;
	dat_format fmt;

	/* Start with unknown file format, and try some heuristics to guess it */
	fmt = DATF_UNKNOWN;
	if (!(gc = g_io_channel_new_file (filename, "r", &err))) {
		g_warning ("Cannot open RC file \"%s\": %s", filename, err -> message);
		g_error_free (err);
		out = NULL;
	} else {
		g_io_channel_set_encoding (gc, NULL, NULL);
		if (g_io_channel_read_line (gc, &line, NULL, NULL, &err) == G_IO_STATUS_NORMAL) {
			g_strstrip (line);
// 			g_print ("--> %s", line);
			if (g_pattern_match_simple ("[*]", line)) {
				/* Our heuristics say this file is in RomCenter format */
				fmt = DATF_ROMCENTER;
			} else {
				/* Let's assume this file is in ClrMamePro format */
				fmt = DATF_CLRMAMEPRO;
			}
			g_free (line);
		} else {
			/* Can't even read a line!? */
			g_assert_not_reached ();
		}
		g_io_channel_unref (gc);		// = close()
	}

	/* Do the actual reading */
	switch (fmt) {
		case DATF_ROMCENTER:
			out = datfile_new_from_filename_rc (filename);
			break;
		case DATF_CLRMAMEPRO:
			out = datfile_new_from_filename_cmp (filename);
			break;
		default:
			out = NULL;
			g_warning ("DAT file \"%s\" has unknown format", filename);
			break;
	}

	return (out);
}
