/***************************************************************************
 *   Copyright (C) 2005 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <sys/stat.h>
#include <glib.h>
#include <glib/gstdio.h>
#include "file.h"

#include "crc32.h"

#ifdef USE_MD5
	#include "md5.h"
#endif

#ifdef USE_SHA1
	#include "sha1.h"
#endif


filez *file_new (char *filename) {
	filez *f;
	struct stat st;

	f = g_new0 (filez, 1);

	f -> filename = g_strdup (filename);
	if (g_stat (filename, &st) < 0) {
		g_warning ("Cannot stat() \"%s\"", filename);
	} else {
		f -> filesize = st.st_size;
	}

	return (f);
}


filez *file_new_archived (char *filename, char *archive, guint32 filesize) {
	filez *f;

	f = g_new0 (filez, 1);

	f -> filename = g_strdup (filename);
	f -> filesize = filesize;
	f -> parent_archive = g_strdup (archive);

	return (f);
}


void file_destroy (filez *f) {
	g_free (f -> filename);
	g_free (f);

	return;
}


void file_compute_sums (filez *f) {
	FILE *fp;
	guint32 crc32;
#ifdef USE_MD5
	MD5_CTX md5ctx;
#endif
#ifdef USE_SHA1
	SHA1_CTX sha1ctx;
#endif
	int bytes;
	unsigned char data[READBUF_SIZE];
#ifdef USE_SHA1
	unsigned char sha1digest[SHA1_DIGESTSIZE];
#endif

	if ((fp = fopen (f -> filename, "r"))) {
		crc32 = 0xffffffff;
#ifdef USE_MD5
		MD5Init (&md5ctx);
#endif
#ifdef USE_SHA1
		SHA1Init (&sha1ctx);
#endif
		while ((bytes = fread (data, 1, READBUF_SIZE, fp)) != 0) {
			crc32 = CrcUpdate (crc32, data, bytes);
#ifdef USE_MD5
			MD5Update (&md5ctx, data, bytes);
#endif
#ifdef USE_SHA1
			SHA1Update (&sha1ctx, data, bytes);		/* WARNING: SHA1Update destroys data! */
#endif
		}
		crc32 ^= 0xffffffff;
#ifdef USE_MD5
		MD5Final (&md5ctx);
#endif
#ifdef USE_SHA1
		SHA1Final (sha1digest, &sha1ctx);
#endif
		/* Save the computed CRC32 in a string the common hex format */
		snprintf (f -> crc32, LEN_CRC32 + 1, "%08lx", crc32);

#ifdef USE_MD5
		/* Save the computed MD5sum in a string the common hex format */
		for (bytes = 0; bytes < LEN_MD5 / 2; bytes++)
			sprintf (f -> md5, "%s%02x", f -> md5, md5ctx.digest[bytes]);
		(f -> md5)[LEN_MD5] = '\0';
#endif

#ifdef USE_SHA1
		/* Save the computed SHA1sum in a string the common hex format */
		for (bytes = 0; bytes < LEN_SHA1 / 2; bytes++)
			sprintf (f -> sha1, "%s%02x", f -> sha1, sha1digest[bytes]);
		(f -> sha1)[LEN_SHA1] = '\0';
#endif

		fclose (fp);
	} else
		/* Cannot open file */;

	return;
}
