/***************************************************************************
 *   Copyright (C) 2006 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib.h>
#include <glib/gstdio.h>
#include "filelist.h"
#include "file.h"
#include "zipfile.h"

GList *add_to_filelist (GList *filelist, char *basedir, gboolean recurse) {
	GDir *dir;
	const gchar *file;
	gchar *fullpath;
	filez *f;
	GList *zl;
	zipfile *zf;

	if ((dir = g_dir_open (basedir, 0, NULL))) {
		g_dir_rewind (dir);
		while ((file = g_dir_read_name (dir))) {
// 			fullpath = g_strconcat (basedir, G_DIR_SEPARATOR_S, file, NULL);
			fullpath = g_build_filename (basedir, file, NULL);
			if (g_file_test (fullpath, G_FILE_TEST_IS_DIR)) {
				if (recurse) {
// 					printf ("--- Recursing into %s\n", file);
					filelist = add_to_filelist (filelist, fullpath, recurse);
				}
			} else if (g_str_has_suffix (file, ".zip") || g_str_has_suffix (file, ".ZIP")) {
				/* Append ZIP file contens to the list */
				g_fprintf (stderr, "Found ZIP file: \"%s\"\n", fullpath);
				zf = zipfile_new (fullpath);
				zl = zipfile_get_filelist (zf);
				zf = zipfile_destroy (zf);

				filelist = g_list_concat (filelist, zl);
				// g_list_free (zl);		/* We don't need freeing, do we? */
			} else {
// 				printf ("Adding %s to filelist\n", file);
				f = file_new (fullpath);
				filelist = g_list_append (filelist, f);
			}
			g_free (fullpath);
		}
		g_dir_close (dir);
// 		printf ("--- End of dir ---\n");
	}

	return (filelist);
}


#define COPY_BUFSIZE 8192

gboolean filecopy (const char *src, const char *dest) {
	gboolean ret;
	FILE *in, *out;
	guint8 buffer[COPY_BUFSIZE];
	size_t bread;

// 	g_debug (_("Copying \"%s\" to \"%s\""), src, dest);

	ret = FALSE;

	if (!(in = g_fopen (src, "rb")))
		g_warning (("Error opening input file"));
	else if (!(out = g_fopen (dest, "wb"))) {
		g_warning (("Error opening output file"));
		fclose (in);
	} else {
		ret = TRUE;
		while (((bread = fread (buffer, 1, COPY_BUFSIZE, in)) > 0) && ret) {
// 			g_debug ("Read %u bytes", bread);
			if (fwrite (buffer, 1, bread, out) < bread) {	// TODO
				g_warning (("Error during copy"));
				ret = FALSE;
			}
		}
		fclose (in);
		fclose (out);
	}

	return (ret);
}


gboolean filemove (const char *src, const char *dest) {
	gboolean out;

	/* Try making a new link first */
	if (link (src, dest) < 0) {
// 		g_debug ("filemove: link() failed, trying filecopy()");
		/* If link() fails, we might be spanning two different filesystems, so we need to do an actual copy */
		out = filecopy (src, dest);
	} else {
		out = TRUE;
	}
	
	/* Remove original link if we had success somehow */
	if (out) {
		if (g_unlink (src) < 0) {
			g_warning ("filemove: cannot unlink old file \"%s\"", src);
			out = FALSE;
		}
	}
	
	return (out);
}
