/***************************************************************************
 *   Copyright (C) 2006 by SukkoPera                                       *
 *   sukkopera@gmail.com                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <fcntl.h>
#include <glib.h>
#include <glib/gstdio.h>

#ifdef unix
# include <unistd.h>
# include <utime.h>
# include <sys/types.h>
# include <sys/stat.h>
#else
# include <direct.h>
# include <io.h>
#endif

#include "mz_zip.h"
#include "mz_unzip.h"

#define CASESENSITIVITY (0)
#define WRITEBUFFERSIZE (8192)
#define MAXFILENAME (256)

#ifdef WIN32
	#define USEWIN32IOAPI
	#include "mz_iowin32.h"
#endif

#include "zipfile.h"
#include "file.h"


/* WARNING: This whole file is a huge hack!!! */


/**
 *
 * @param f name of file to get info on
 * @param tmzip return value: access, modific. and creation times
 * @param dt dostime
 * @return
 */
uLong filetime (char *f,tm_zip *tmzip, uLong *dt) {
	int ret = 0;
#ifdef WIN32
	FILETIME ftLocal;
	HANDLE hFind;
	WIN32_FIND_DATA  ff32;
	
	hFind = FindFirstFile(f,&ff32);
	if (hFind != INVALID_HANDLE_VALUE) {
		FileTimeToLocalFileTime(&(ff32.ftLastWriteTime),&ftLocal);
		FileTimeToDosDateTime(&ftLocal,((LPWORD)dt)+1,((LPWORD)dt)+0);
		FindClose(hFind);
		ret = 1;
	}
#elif defined (unix)
	struct stat s;        /* results of stat() */
	struct tm* filedate;
	time_t tm_t = 0;
	
	if (strcmp (f, "-") != 0) {
		char name[MAXFILENAME + 1];
		int len = strlen (f);
		if (len > MAXFILENAME)
			len = MAXFILENAME;
	
		strncpy (name, f, MAXFILENAME - 1);
		/* strncpy doesnt append the trailing NULL, of the string is too long. */
		name[MAXFILENAME] = '\0';
	
		if (name[len - 1] == '/')
			name[len - 1] = '\0';
		/* not all systems allow stat'ing a file with / appended */
		if (stat (name, &s)==0) {
			tm_t = s.st_mtime;
			ret = 1;
		}
	}
	filedate = localtime (&tm_t);
	
	tmzip->tm_sec  = filedate->tm_sec;
	tmzip->tm_min  = filedate->tm_min;
	tmzip->tm_hour = filedate->tm_hour;
	tmzip->tm_mday = filedate->tm_mday;
	tmzip->tm_mon  = filedate->tm_mon ;
	tmzip->tm_year = filedate->tm_year;
#endif
	return (ret);
}

/* change_file_date : change the date/time of a file
    filename : the filename of the file where date/time must be modified
    dosdate : the new date at the MSDos format (4 bytes)
    tmu_date : the SAME new date at the tm_unz format */
void change_file_date (const char *filename, uLong dosdate, tm_unz tmu_date) {
#ifdef WIN32
  HANDLE hFile;
  FILETIME ftm,ftLocal,ftCreate,ftLastAcc,ftLastWrite;

  hFile = CreateFile(filename,GENERIC_READ | GENERIC_WRITE,
		     0,NULL,OPEN_EXISTING,0,NULL);
  GetFileTime(hFile,&ftCreate,&ftLastAcc,&ftLastWrite);
  DosDateTimeToFileTime((WORD)(dosdate>>16),(WORD)dosdate,&ftLocal);
  LocalFileTimeToFileTime(&ftLocal,&ftm);
  SetFileTime(hFile,&ftm,&ftLastAcc,&ftm);
  CloseHandle(hFile);
#else
#ifdef unix
  struct utimbuf ut;
  struct tm newdate;
  newdate.tm_sec = tmu_date.tm_sec;
  newdate.tm_min=tmu_date.tm_min;
  newdate.tm_hour=tmu_date.tm_hour;
  newdate.tm_mday=tmu_date.tm_mday;
  newdate.tm_mon=tmu_date.tm_mon;
  if (tmu_date.tm_year > 1900)
	  newdate.tm_year=tmu_date.tm_year - 1900;
  else
	  newdate.tm_year=tmu_date.tm_year ;
  newdate.tm_isdst=-1;

  ut.actime=ut.modtime=mktime(&newdate);
  utime(filename,&ut);
#endif
#endif
}


static int do_extract_currentfile (unzFile uf, const char *destname, const gboolean opt_extract_without_path,
				   					gboolean opt_overwrite, const char *password) {
	char filename_inzip[256], *filename_withoutpath, *p;
	int err = UNZ_OK;
	FILE *fout = NULL;
	void *buf;
	uInt size_buf;

	unz_file_info file_info;
	err = unzGetCurrentFileInfo (uf, &file_info, filename_inzip, sizeof (filename_inzip), NULL, 0, NULL, 0);

	if (err != UNZ_OK) {
		g_warning ("Error %d with zipfile in unzGetCurrentFileInfo", err);
		return err;
	}

	size_buf = WRITEBUFFERSIZE;
	buf = g_new (void *, size_buf);
	g_assert (buf);

	p = filename_withoutpath = filename_inzip;
	while ((*p) != '\0') {
		if (((*p)=='/') || ((*p)=='\\'))
			filename_withoutpath = p+1;
		p++;
	}
// 	g_debug ("fwp: %s|fiz: %s", filename_withoutpath, filename_inzip);

	if ((*filename_withoutpath) == '\0') {
		/* FIXME I think this creates the wrong directory! */
		if (!opt_extract_without_path) {
			g_message ("Creating directory: %s", filename_inzip);
			g_mkdir (filename_inzip, 0755);
		}
	} else {
		const char *write_filename;

		if (destname) {
			/* This forces opt_extract_without_path, of course */
			write_filename = destname;
		} else if (!opt_extract_without_path)
			write_filename = filename_inzip;
		else
			write_filename = filename_withoutpath;

		err = unzOpenCurrentFilePassword (uf, password);
		if (err != UNZ_OK)
			g_warning ("Error %d with zipfile in unzOpenCurrentFilePassword", err);
		else if (!opt_overwrite && g_file_test (write_filename, G_FILE_TEST_EXISTS))
			g_message ("Skipping %s because it already exists and not overwriting", write_filename);
		else {
			fout = fopen (write_filename, "wb");

			/* some zipfiles don't contain directory alone before file */
			if (!fout && !opt_extract_without_path &&
						(filename_withoutpath != (char *) filename_inzip)) {
				char c = *(filename_withoutpath - 1);
				*(filename_withoutpath - 1) = '\0';
				g_mkdir (write_filename, 0755);
				*(filename_withoutpath - 1) = c;
				fout = fopen (write_filename, "wb");
			}

			if (!fout)
				g_warning ("Error opening %s", write_filename);
			else {
// 				g_message ("Extracting \"%s\" to \"%s\"", filename_inzip, write_filename);

				do {
					err = unzReadCurrentFile (uf,buf,size_buf);
					if (err < 0) {
						g_warning ("Error %d with zipfile in unzReadCurrentFile", err);
						break;
					} else if (err > 0) {
						if (fwrite (buf, err, 1, fout) != 1) {
							g_warning ("Error in writing extracted file");
							err = UNZ_ERRNO;
							break;
						}
					}
				} while (err > 0);

				fclose (fout);

				if (err == 0)
					change_file_date (write_filename, file_info.dosDate, file_info.tmu_date);
			}
		}

		if (err == UNZ_OK) {
			err = unzCloseCurrentFile (uf);		/* I think this will check the file CRC against the stored one */
			if (err != UNZ_OK)
				g_warning ("Error %d with zipfile in unzCloseCurrentFile", err);
		} else
			unzCloseCurrentFile (uf); /* don't lose the error */
	}

	g_free (buf);
	return (err);
}


int zipfile_extract_all (zipfile *zf, char *destdir, gboolean nofullpath, gboolean overwrite) {
	uLong i;
	unz_global_info gi;
	int err;
	unzFile uf;

#ifdef USEWIN32IOAPI
	zlib_filefunc_def ffunc;
#endif

#ifdef USEWIN32IOAPI
	fill_win32_filefunc (&ffunc);
	uf = unzOpen2 (zf -> filename, &ffunc);
#else
	uf = unzOpen (zf -> filename);
#endif

	if (!uf) {
		fprintf (stderr, "Error opening zip file\n");
	} else if ((err = unzGetGlobalInfo (uf, &gi)) != UNZ_OK) {
		fprintf (stderr, "error %d with zipfile in unzGetGlobalInfo \n", err);
	} else {
		for (i = 0; i < gi.number_entry; i++) {
			if (do_extract_currentfile (uf, NULL, nofullpath, overwrite, zf -> password) != UNZ_OK)
				break;
		
			if ((i + 1) < gi.number_entry) {
				err = unzGoToNextFile (uf);
				if (err != UNZ_OK) {
					printf ("error %d with zipfile in unzGoToNextFile\n", err);
					break;
				}
			}
		}
	
		unzCloseCurrentFile (uf);
	}

	return (0);
}


/* FIXME Return value */
int zipfile_extract_one (zipfile *zf, char *file_to_extract, char *destfile, gboolean nofullpath, gboolean overwrite) {
	uLong i;
	unz_global_info gi;
	int err;
	unzFile uf;
	unz_file_info file_info;
	gboolean found, out;
	char filename_inzip[256];

#ifdef USEWIN32IOAPI
	zlib_filefunc_def ffunc;
#endif

#ifdef USEWIN32IOAPI
	fill_win32_filefunc (&ffunc);
	uf = unzOpen2 (zf -> filename, &ffunc);
#else
	uf = unzOpen (zf -> filename);
#endif

	if (!uf) {
		fprintf (stderr, "Error opening zip file\n");
	} else if ((err = unzGetGlobalInfo (uf, &gi)) != UNZ_OK) {
		fprintf (stderr, "error %d with zipfile in unzGetGlobalInfo \n", err);
	} else {
		found = FALSE;
		for (i = 0; !found && i < gi.number_entry; i++) {
			err = unzGetCurrentFileInfo (uf, &file_info, filename_inzip, sizeof (filename_inzip), NULL, 0, NULL, 0);
			if (err != UNZ_OK) {
				printf ("error %d with zipfile in unzGetCurrentFileInfo\n", err);
				break;
			}

			if (strcmp (filename_inzip, file_to_extract) == 0)
				found = TRUE;
			else {
				/* On with next file */
				if ((i + 1) < gi.number_entry) {
					err = unzGoToNextFile (uf);
					if (err != UNZ_OK) {
						printf ("error %d with zipfile in unzGoToNextFile\n", err);
						break;
					}
				}
			}
		}

		if (found) {
			if (do_extract_currentfile (uf, destfile, nofullpath, overwrite, zf -> password) != UNZ_OK)
				out = FALSE;
		} else {
			g_assert_not_reached ();
		}
	
		unzCloseCurrentFile (uf);
	}

	return (0);
}


zipfile *zipfile_new (char *filename) {
	zipfile *zf;

	zf = g_new0 (zipfile, 1);
	zf -> filename = g_strdup (filename);
	zf -> password = NULL;
	zf -> compress_level = Z_DEFAULT_COMPRESSION;

	return (zf);
}


zipfile *zipfile_destroy (zipfile *zf) {
	/* If the user forgot to call zipfile_end_add(), do it for him */
	if (zf -> zf_int)
			zipfile_end_add (zf);		/* DOES NOT WORK CORRECTLY ATM */
	g_free (zf -> filename);
	g_free (zf -> password);
	g_free (zf);
	
	return (NULL);
}
	

/* TODO: turn to g_print() */
void zipfile_dump (zipfile *zf) {
	uLong i;
	unz_global_info gi;
	int err;
	unzFile uf;

#ifdef USEWIN32IOAPI
	zlib_filefunc_def ffunc;
#endif

#ifdef USEWIN32IOAPI
	fill_win32_filefunc (&ffunc);
	uf = unzOpen2 (zf -> filename, &ffunc);
#else
	uf = unzOpen (zf -> filename);
#endif

	if (!uf) {
		fprintf (stderr, "Error opening zip file\n");
	} else {
		err = unzGetGlobalInfo (uf,&gi);
		if (err!=UNZ_OK)
			printf("error %d with zipfile in unzGetGlobalInfo \n",err);
		printf(" Length  Method   Size  Ratio   Date    Time   CRC-32     Name\n");
		printf(" ------  ------   ----  -----   ----    ----   ------     ----\n");
		for (i=0;i<gi.number_entry;i++)
		{
			char filename_inzip[256];
			unz_file_info file_info;
			uLong ratio=0;
			const char *string_method;
			char charCrypt=' ';
			err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);
			if (err!=UNZ_OK)
			{
				printf("error %d with zipfile in unzGetCurrentFileInfo\n",err);
				break;
			}
			if (file_info.uncompressed_size>0)
				ratio = (file_info.compressed_size*100)/file_info.uncompressed_size;
	
			/* display a '*' if the file is crypted */
			if ((file_info.flag & 1) != 0)
				charCrypt='*';
	
			if (file_info.compression_method==0)
				string_method="Stored";
			else
				if (file_info.compression_method==Z_DEFLATED)
			{
				uInt iLevel=(uInt)((file_info.flag & 0x6)/2);
				if (iLevel==0)
					string_method="Defl:N";
				else if (iLevel==1)
					string_method="Defl:X";
				else if ((iLevel==2) || (iLevel==3))
					string_method="Defl:F"; /* 2:fast , 3 : extra fast*/
			}
			else
				string_method="Unkn. ";
	
			printf("%7lu  %6s%c%7lu %3lu%%  %2.2lu-%2.2lu-%2.2lu  %2.2lu:%2.2lu  %8.8lx   %s\n",
			file_info.uncompressed_size,string_method,
			charCrypt,
			file_info.compressed_size,
			ratio,
			(uLong)file_info.tmu_date.tm_mon + 1,
			(uLong)file_info.tmu_date.tm_mday,
			(uLong)file_info.tmu_date.tm_year % 100,
			(uLong)file_info.tmu_date.tm_hour,(uLong)file_info.tmu_date.tm_min,
			(uLong)file_info.crc,filename_inzip);
			if ((i+1)<gi.number_entry)
			{
				err = unzGoToNextFile(uf);
				if (err!=UNZ_OK)
				{
					printf("error %d with zipfile in unzGoToNextFile\n",err);
					break;
				}
			}
		}
	
		unzCloseCurrentFile(uf);
	}

	return;
}



/* calculate the CRC32 of a file,
   because to encrypt a file, we need known the CRC32 of the file before */
static int getFileCrc (const char* filenameinzip, void *buf, unsigned long size_buf, unsigned long *result_crc) {
	unsigned long calculate_crc=0;
	int err=ZIP_OK;
	FILE * fin = fopen(filenameinzip,"rb");
	unsigned long size_read = 0;
	unsigned long total_read = 0;
	if (fin==NULL)
	{
		err = ZIP_ERRNO;
	}

	if (err == ZIP_OK)
		do
	{
		err = ZIP_OK;
		size_read = (int)fread(buf,1,size_buf,fin);
		if (size_read < size_buf)
			if (feof(fin)==0)
		{
			printf("error in reading %s\n",filenameinzip);
			err = ZIP_ERRNO;
		}

		if (size_read>0)
			calculate_crc = crc32(calculate_crc,buf,size_read);
		total_read += size_read;

	} while ((err == ZIP_OK) && (size_read>0));

	if (fin)
		fclose(fin);

	*result_crc=calculate_crc;
	printf ("file %s crc %lx\n", filenameinzip, calculate_crc);
	return err;
}


gboolean zipfile_begin_add (zipfile *zf) {
	int overwrite_flag;

	g_assert (zf);
	
	/* Quite crap... */
	if (g_file_test (zf -> filename, G_FILE_TEST_EXISTS))
		overwrite_flag = 2;
	else
		overwrite_flag = 0;

	zf -> bufsize = WRITEBUFFERSIZE;
	zf -> buffer = g_new (void *, zf -> bufsize);
	g_assert (zf -> buffer);

#ifdef USEWIN32IOAPI
	zlib_filefunc_def ffunc;
	fill_win32_filefunc (&ffunc);
	zf -> zf_int = zipOpen2 (zf -> filename, overwrite_flag, NULL, &ffunc);
#else
        zf -> zf_int = zipOpen (zf -> filename, overwrite_flag);
#endif

	return (zf -> zf_int != NULL);
}


int zipfile_add (zipfile *zf, char *srcfile, char *filenameinzip, char *filecomment) {
	FILE *fin;
	int err, size_read;
	zip_fileinfo zi;
	unsigned long crcFile;

	g_assert (zf && zf -> zf_int);

	memset (&zi, 0, sizeof (zip_fileinfo));			/* Very important! */
	filetime (filenameinzip, &zi.tmz_date, &zi.dosDate);

	crcFile = 0;
	if (zf -> password)
		err = getFileCrc (filenameinzip, zf -> buffer, zf -> bufsize, &crcFile);

	err = zipOpenNewFileInZip3 (zf -> zf_int, filenameinzip, &zi,
			NULL, 0, NULL, 0, filecomment,
			(zf -> compress_level != 0) ? Z_DEFLATED : 0, zf -> compress_level, 0,
			-MAX_WBITS, DEF_MEM_LEVEL, Z_DEFAULT_STRATEGY, zf -> password, crcFile);

	if (err != ZIP_OK)
		g_warning ("Error opening \"%s\" in zipfile", filenameinzip);
	else {
		fin = fopen (srcfile, "rb");
		if (fin == NULL) {
			err = ZIP_ERRNO;
			g_warning ("Error in opening \"%s\" for reading", srcfile);
		}

		if (err == ZIP_OK)
			do {
				err = ZIP_OK;
				size_read = (int) fread (zf -> buffer, 1, zf -> bufsize, fin);
				if (size_read < zf -> bufsize) {
					if (!feof (fin)) {
						g_warning ("Error in reading %s", filenameinzip);
						err = ZIP_ERRNO;
					}
				}
		
				if (size_read > 0) {
					err = zipWriteInFileInZip (zf -> zf_int, zf -> buffer, size_read);
					if (err < 0) {
						g_warning ("Error in writing %s in the zipfile", filenameinzip);
					}
				}
			} while ((err == ZIP_OK) && (size_read > 0));
	
		if (fin)
			fclose (fin);
	
		if (err < 0)
			err = ZIP_ERRNO;
		else {
			err = zipCloseFileInZip (zf -> zf_int);
			if (err != ZIP_OK)
				g_warning ("Error in closing %s in the zipfile", filenameinzip);
		}
	}
	
	return (err == ZIP_OK);
}


gboolean zipfile_end_add (zipfile *zf) {
	int err;

	g_assert (zf);
	
	g_free (zf -> buffer);
	zf -> bufsize = 0;
	
	err = zipClose (zf -> zf_int, NULL);
	zf -> zf_int = NULL;
	if (err != ZIP_OK)
		g_warning ("Error in closing %s", zf -> filename);

	return (err == ZIP_OK);
}


GList *zipfile_get_filelist (zipfile *zf) {
	uLong i;
	unz_global_info gi;
	int err;
	unzFile uf;
	char filename_inzip[256], *tmp;
	filez *f;
	GList *out;

#ifdef USEWIN32IOAPI
	zlib_filefunc_def ffunc;
#endif

	/* Init GList */
	out = NULL;

#ifdef USEWIN32IOAPI
	fill_win32_filefunc (&ffunc);
	uf = unzOpen2 (zf -> filename, &ffunc);
#else
	uf = unzOpen (zf -> filename);
#endif

	if (!uf) {
		fprintf (stderr, "Error opening zip file\n");
	} else {
// 		g_debug ("--> ZIP file: %s", zf -> filename);
		err = unzGetGlobalInfo (uf, &gi);
		if (err != UNZ_OK)
			printf ("error %d with zipfile in unzGetGlobalInfo \n", err);
		
		for (i = 0; i < gi.number_entry; i++) {
			unz_file_info file_info;
			err = unzGetCurrentFileInfo (uf, &file_info, filename_inzip, sizeof (filename_inzip), NULL, 0, NULL, 0);
			if (err != UNZ_OK) {
				printf ("error %d with zipfile in unzGetCurrentFileInfo\n", err);
				break;
			}

// 			g_debug ("|---> %s (%08lx)", filename_inzip, file_info.crc);
			tmp = g_build_filename (zf -> filename, filename_inzip, NULL);
			f = file_new_archived (tmp, zf -> filename, file_info.uncompressed_size);
			sprintf (f -> crc32, "%08lx", (uLong) file_info.crc);
			g_free (tmp);
			out = g_list_append (out, f);
			
			if ((i + 1) < gi.number_entry) {
				err = unzGoToNextFile (uf);
				if (err != UNZ_OK) {
					printf ("error %d with zipfile in unzGoToNextFile\n", err);
					break;
				}
			}
		}
	
		unzCloseCurrentFile(uf);
	}

	return (out);
}
